# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Layout & transform inference convenience utils."""

from collections.abc import Sequence
from functools import partial
from typing import Union

from jax._src.lib import mosaic_gpu_dialect as mgpu
from jax._src.lib.mlir import ir

from . import fragmented_array as fa
from . import tcgen05
from . import utils

MlirOperation = Union[ir.Operation, ir.OpView]


def in_layouts(op: MlirOperation) -> Sequence[ir.Attribute]:
  """Returns the in_layouts attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an in_layouts attribute.
  """
  return _array_attr(op, "in_layouts")


def out_layouts(op: MlirOperation) -> Sequence[ir.Attribute]:
  """Returns the out_layouts attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an out_layouts attribute.
  """
  return _array_attr(op, "out_layouts")


def in_transforms(op: MlirOperation) -> Sequence[ir.ArrayAttr]:
  """Returns the in_transforms attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an in_transforms attribute.
  """
  return _array_attr(op, "in_transforms")  # pyrefly: ignore[bad-return]


def out_transforms(op: MlirOperation) -> Sequence[ir.ArrayAttr]:
  """Returns the out_transforms attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an out_transforms attribute.
  """
  return _array_attr(op, "out_transforms")  # pyrefly: ignore[bad-return]


def in_tmem_layouts(op: MlirOperation) -> Sequence[ir.Attribute]:
  """Returns the in_tmem_layouts attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an in_tmem_layouts attribute.
  """
  return _array_attr(op, "in_tmem_layouts")


def out_tmem_layouts(op: MlirOperation) -> Sequence[ir.Attribute]:
  """Returns the out_tmem_layouts attribute of the given operation.

  Raises:
    ValueError: If the operation does not have an out_tmem_layouts attribute.
  """
  return _array_attr(op, "out_tmem_layouts")


def _array_attr(op: MlirOperation, name: str) -> Sequence[ir.Attribute]:
  try:
    result = op.attributes[name]
  except KeyError:
    raise ValueError(f"{op} does not have an {name} attribute") from None
  if not isinstance(result, ir.ArrayAttr):
    raise TypeError(f"{op} has {name} of an unexpected type: {result}")
  return result  # pyrefly: ignore[bad-return]


def should_have_in_tmem_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation operands should be assigned a TMEM layout."""
  return any(
      isinstance(v.type, ir.MemRefType) and utils.is_tmem_ref(v)
      for v in op.operands
  )


def should_have_out_tmem_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation results should be assigned a TMEM layout."""
  return any(
      isinstance(v.type, ir.MemRefType) and utils.is_tmem_ref(v)
      for v in op.results
  )


def should_have_tmem_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation should be assigned a TMEM layout."""
  return should_have_in_tmem_layout(op) or should_have_out_tmem_layout(op)


def should_have_in_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation operands should be assigned a layout."""
  return any(isinstance(v.type, ir.VectorType) for v in op.operands)


def should_have_out_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation results should be assigned a layout."""
  return any(isinstance(v.type, ir.VectorType) for v in op.results)


def should_have_layout(op: MlirOperation) -> bool:
  """Returns 'true' if the operation should be assigned a layout."""
  return should_have_in_layout(op) or should_have_out_layout(op)


def has_in_layouts_set(op: MlirOperation) -> bool:
  return "in_layouts" in op.attributes


def has_out_layouts_set(op: MlirOperation) -> bool:
  return "out_layouts" in op.attributes


def has_any_layout_set(op: MlirOperation) -> bool:
  return has_in_layouts_set(op) or has_out_layouts_set(op)


def has_out_transforms_set(op: MlirOperation) -> bool:
  return "out_transforms" in op.attributes


def attr_element(
    attr_name: str, op: MlirOperation, index: int
) -> ir.Attribute | None:
  """Returns `op.attributes[attr_name][index]` if it exists, otherwise None.

  If `op.attributes[attr_name]` exists, then `index` must be a valid index into
  the attribute array.
  """
  if attr_name not in op.attributes:
    return None
  attr = op.attributes[attr_name]
  if not attr:
    return None
  assert isinstance(attr, ir.ArrayAttr)
  return attr[index]


def _in_attr_for_operand(
    op: MlirOperation,
    operand: ir.Value,
    attr_name: str,
) -> ir.Attribute | None:
  if attr_name == "in_layouts":
    predicate = lambda v: isinstance(v.type, ir.VectorType)
  elif attr_name == "in_transforms":
    predicate = is_transformable_smem_memref
  elif attr_name == "in_tmem_layouts":
    predicate = (
        lambda v: isinstance(v.type, ir.MemRefType)
        and ir.MemRefType(v.type).memory_space == utils.tmem()
    )
  else:
    raise ValueError(f"Unknown attribute: {attr_name}")

  operand_number = [o for o in op.operands if predicate(o)].index(operand)

  return attr_element(attr_name, op, operand_number)


in_layout_for_operand = partial(
    _in_attr_for_operand, attr_name="in_layouts"
)
in_tmem_layout_for_operand = partial(
    _in_attr_for_operand, attr_name="in_tmem_layouts"
)
in_transforms_for_operand = partial(
    _in_attr_for_operand, attr_name="in_transforms"
)


def should_have_in_transforms(op: ir.OpView) -> bool:
  """Returns 'True' if the operation should be assigned in transforms."""
  return any(map(is_transformable_smem_memref, op.operands))


def should_have_out_transforms(op: ir.OpView) -> bool:
  """Returns 'True' if the operation should be assigned out transforms."""
  return any(map(is_transformable_smem_memref, op.results))


def should_have_transforms(op: ir.OpView) -> bool:
  """Returns 'True' if the operation should be assigned in/out transforms."""
  return should_have_in_transforms(op) or should_have_out_transforms(op)


def is_transformable_smem_memref(v: ir.Value) -> bool:
  """Whether the value is a memref in SMEM on which transforms should be applied."""
  return (
      isinstance(v.type, ir.MemRefType)
      # barriers have no business being transformed
      and not isinstance(v.type.element_type, mgpu.BarrierType)
      and utils.is_smem_ref(v)
  )


def is_mma_layout(layout: fa.FragmentedLayout) -> bool:
  if not isinstance(layout, fa.TiledLayout):
    return False
  if layout in {
      fa.WGMMA_LAYOUT,
      fa.WGMMA_LAYOUT_ACC_32BIT,
      fa.WGMMA_LAYOUT_UPCAST_2X,
      fa.WGMMA_LAYOUT_UPCAST_4X,
      fa.WGMMA_TRANSPOSED_LAYOUT,
      fa.WGMMA_LAYOUT_8BIT,
      fa.TCGEN05_LAYOUT,
      fa.TCGEN05_TRANSPOSED_LAYOUT,
  }:
    return True
  if len(layout.tiling.tiles[0]) != 2:
    return False
  columns = layout.tiling.tiles[0][1]
  return columns % 16 == 0 and (
      layout == tcgen05.fa_m64_collective_layout(columns)
  )



# Copyright 2026 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Mapping, Sequence
import dataclasses
import functools
import math
from typing import Any

import jax
from jax import lax
from jax._src import core as jax_core
from jax._src import source_info_util
from jax._src.pallas import primitives
from jax._src.pallas.mosaic.interpret import utils as interpret_utils
from jax._src.pallas.mosaic_gpu import core as mosaic_gpu_core
from jax._src.pallas.mosaic_gpu import primitives as gpu_primitives
from jax._src.pallas.mosaic_gpu.interpret import gpu_callbacks
from jax._src.pallas.mosaic_gpu.interpret.params import InterpretGPUParams
from jax._src.state import indexing
from jax._src.state import primitives as state_primitives
from jax._src.state import types as state_types
from jax._src.util import (safe_zip, split_list)
from jax.experimental.pallas import mosaic_gpu as plgpu
import jax.numpy as jnp


@dataclasses.dataclass(init=False, frozen=True)
class DeviceInfo:
  """Information about the device that is being interpreted."""

  # The indices along each axis of the device being interpreted.
  axis_indices: Mapping[jax_core.AxisName, int]
  # The size of each axis in the mesh of all (SMPD) devices.
  axis_sizes: Mapping[jax_core.AxisName, int]

  def __init__(self):
    # Since this class is frozen, we must use `object.__setattr__` to set the
    # attributes.
    object.__setattr__(self, "axis_sizes", jax_core.get_axis_env().axis_sizes)
    object.__setattr__(
        self,
        "axis_indices",
        {k: lax.axis_index(k) for k in self.axis_sizes.keys()},
    )

  @functools.cached_property
  def device_id(self) -> int:
    """Computes the logical ID of the device being interpreted."""
    return interpret_utils.device_coords_to_logical_id(
        tuple(self.axis_indices.values()), self.axis_sizes, self.axis_indices
    )

  @functools.cached_property
  def num_devices(self) -> int:
    """Computes the number of (SPMD) devices."""
    return math.prod(self.axis_sizes.values())


def _raise_if_unsupported_memory_space(
    space: mosaic_gpu_core.MemorySpace | None,
):
  # TODO(nrink): Support more memory spaces.
  if space is not None and space not in [
      mosaic_gpu_core.MemorySpace.GMEM,
      mosaic_gpu_core.MemorySpace.SMEM,
  ]:
    raise NotImplementedError(f"Unsupported memory space: {space}")


# TODO(nrink): Try unifying this function with `_extract_barrier_slice_base`
# from `jax._src.pallas.mosaic_gpu.primitives`.
def _get_index_for_barrier_allocation_key(
    transforms_treedef, transforms_leaves,
) -> indexing.DimIndexer | None:
  # TODO(nrink): The working out of `transforms` and the returned index below
  # may need tidying up. Specifically, GPU interpret mode should correctly
  # support legal ways to index into barriers. (Here, 'legal' is to be read as
  # 'allowed by the Pallas GPU semantics'.)
  transforms = jax.tree.unflatten(transforms_treedef, transforms_leaves)

  if not transforms:
    return None
  if not hasattr(transforms, "__len__") or len(transforms) != 1:
    raise NotImplementedError(
        f"Indexing barrier with {transforms} not supported in GPU interpret"
        " mode"
    )
  if not isinstance(transforms[0], indexing.NDIndexer):
    raise ValueError(f"Expected an `NDIndexer`, but got {transforms[0]}")
  if len(transforms[0].indices) != 1:
    raise ValueError(
        f"Expected a singleton index, but got {transforms[0].indices}"
    )
  return transforms[0].indices[0]


def _get_barrier_allocation_key_from_inval(
    inval, transforms_treedef, transforms_leaves
) -> jax.Array:
  # `inval` is expected to correspond to a barrier. Since we are interpreting,
  # `inval` will in fact contain the allocation key (which is a Jax array) for
  # the barrier.
  allocation_key_as_array = inval

  # Assert to check internal consistency: `allocation_key_as_array` should be
  # a 2-dim array (and the size of the first dimension equals the
  # `num_barriers` parameter from when the barrier was allocated).
  assert len(allocation_key_as_array.shape) == 2
  num_barriers = allocation_key_as_array.shape[0]

  index = _get_index_for_barrier_allocation_key(
      transforms_treedef, transforms_leaves
  )

  if index is None:
    if num_barriers != 1:
      raise ValueError(
          "Attempting to operate on barrier without indexing, but"
          f" `num_barriers = {num_barriers}`"
      )
    return allocation_key_as_array[0]
  else:
    return allocation_key_as_array[index]


_SENTINEL = jnp.inf


@dataclasses.dataclass(frozen=True, kw_only=True)
class JaxprInterpreter:
  """Interprets a jaxpr by replacing memory operations with (GPU) callbacks."""

  grid_point_coords: tuple[int, ...]
  cluster_dims: tuple[int, ...]

  # The (flat) thread ID for the thread that this interpreter instance is
  # executing. On each simulated GPU, and for each grid point, we execute the
  # following number of threads concurrently (with concurrent instances of
  # `JaxprInterpreter`):
  #
  #   num_threads = math.prod(self.cluster_dims) * self.num_threads_per_block
  #
  # Hence, `thread_id` must be a number in [0, num_threads - 1].
  #
  # The `thread_id` maps to a thread's coordinates along the cluster and thread
  # axes as follows:
  #
  #   thread_coord_tuple = (*self.cluster_coords, self.thread_id_in_block),
  #
  # where the `self.thread_id_in_block` is the minor-most coordinate, and
  # `self.cluster_coords` are in major-to-minor order.
  thread_id: int

  mesh: plgpu.Mesh | None
  device_info: DeviceInfo
  compiler_params: Mapping[str, Any]
  interpret_params: InterpretGPUParams

  @functools.cached_property
  def num_threads_per_block(self) -> int:
    if self.mesh is None or self.mesh.num_threads is None:
      return 1
    else:
      return int(self.mesh.num_threads)

  @functools.cached_property
  def thread_id_in_block(self) -> int:
    return self.thread_id % self.num_threads_per_block

  @functools.cached_property
  def cluster_coords(self) -> tuple[int, ...]:
    thread_id = self.thread_id // self.num_threads_per_block
    return interpret_utils.get_indices(self.cluster_dims, thread_id)

  def _interpret_axis_index_p(self, eqn):
    assert eqn.primitive is lax.axis_index_p
    axis_name = eqn.params["axis_name"]
    if self.mesh is not None:
      if axis_name == self.mesh.thread_name:
        return jnp.int32(self.thread_id_in_block)
      elif axis_name in self.mesh.cluster_names:
        return jnp.int32(
            self.cluster_coords[self.mesh.cluster_names.index(axis_name)]
        )
      elif axis_name in self.mesh.grid_names:
        return jnp.int32(
            self.grid_point_coords[self.mesh.grid_names.index(axis_name)]
        )

    if axis_name in self.device_info.axis_indices:
      return jnp.int32(self.device_info.axis_indices[axis_name])

    raise ValueError(
          f"Unable to determine axis index for axis name {axis_name}"
      )

  def _interpret_get_p(self, eqn, get_invals: Callable[[], Sequence[Any]]):
    assert eqn.primitive is state_primitives.get_p
    assert isinstance(eqn.outvars[0].aval, jax_core.ShapedArray)
    invals = get_invals()
    return gpu_callbacks.call_get(
        result_shape_and_dtype=eqn.outvars[0].aval,
        device_id=self.device_info.device_id,
        thread_id=self.thread_id,
        allocation_key=invals[0],
        transforms=jax.tree.unflatten(eqn.params["tree"], invals[1:]),
        source_info=eqn.source_info,
    )

  def _interpret_swap_p(self, eqn, get_invals: Callable[[], Sequence[Any]]):
    assert eqn.primitive is state_primitives.swap_p
    assert isinstance(eqn.outvars[0].aval, jax_core.ShapedArray)
    invals = get_invals()
    return gpu_callbacks.call_swap(
        result_shape_and_dtype=eqn.outvars[0].aval,
        device_id=self.device_info.device_id,
        thread_id=self.thread_id,
        allocation_key=invals[0],
        transforms=jax.tree.unflatten(eqn.params["tree"], invals[2:]),
        val=invals[1],
        mask=None,
        source_info=eqn.source_info,
    )

  def _interpret_run_scoped_p(
      self, eqn, get_invals: Callable[[], Sequence[Any]]
  ):
    assert eqn.primitive is primitives.run_scoped_p

    def _allocate_for_aval(aval, same_allocations_for_all_threads: bool):
      _raise_if_unsupported_memory_space(aval.memory_space)
      match aval:
        case state_types.AbstractRef(
            inner_aval=inner, memory_space=memory_space, kind=_
        ):
          match inner:
            case jax_core.ShapedArray(shape=shape, dtype=dtype):
              if isinstance(dtype, mosaic_gpu_core.BarrierType):
                # Allocating a barrier is meaningful only if the barrier is
                # shared between all threads. Hence we assert on
                # `same_allocations_for_all_threads`.
                assert same_allocations_for_all_threads
                assert len(shape) == 1
                return gpu_callbacks.call_allocate_barriers(
                    device_id=self.device_info.device_id,
                    thread_id=self.thread_id,
                    num_arrivals=dtype.num_arrivals,
                    num_barriers=shape[0],
                    ref_count=self.num_threads_per_block,
                    source_info=eqn.source_info,
                )
              else:
                memory_space_idx = gpu_callbacks.get_memory_space_idx(
                    memory_space
                )
                allocation_request = (
                    gpu_callbacks.make_allocation_request_array(
                        device_id=self.device_info.device_id,
                        memory_space_id=memory_space_idx,
                        thread_id=(
                            0
                            if same_allocations_for_all_threads
                            else self.thread_id
                        ),
                        initial_ref_count=(
                            self.num_threads_per_block
                            if same_allocations_for_all_threads
                            else 1
                        ),
                    )
                )
              return gpu_callbacks.call_allocate_buffer(
                  self.device_info.device_id,
                  self.thread_id,
                  allocation_request,
                  interpret_utils.get_uninitialized_array(
                      shape, dtype, self.interpret_params.uninitialized_memory),
                  source_info=eqn.source_info,
              )
            case _:
              raise ValueError(f"Unsupported inner aval: {inner}")

    def _deallocate_for_aval(allocation, aval):
      match aval:
        case state_types.AbstractRef(inner_aval=inner, memory_space=_, kind=_):
          match inner:
            case jax_core.ShapedArray(shape=_, dtype=dtype):
              if isinstance(dtype, mosaic_gpu_core.BarrierType):
                gpu_callbacks.call_deallocate_barrier(
                    device_id=self.device_info.device_id,
                    thread_id=self.thread_id,
                    allocation_key=allocation,
                    source_info=eqn.source_info,
                )
              else:
                _raise_if_unsupported_memory_space(aval.memory_space)
                gpu_callbacks.call_deallocate_buffer(
                    self.device_info.device_id,
                    self.thread_id,
                    allocation,
                    source_info=eqn.source_info,
                )

              # TODO(nrink): For sempahores, check that they have value zero at
              # the end of their lifetimes. (If semaphores are never explicitly
              # deallocated, this check could take place at the end of kernel
              # interpretation.)
            case _:
              assert False, (
                  f"Unsupported inner aval: {inner} (should have been"
                  " caught before)"
              )

    assert eqn.primitive is primitives.run_scoped_p
    collective_axes = eqn.params["collective_axes"]
    # Note that on GPU, `SMEM` buffers and barriers can only be allocated
    # collectively (i.e. corresponding to `same_allocations=True`). In the
    # interpreter we are a little more lenient and allow non-collective
    # allocations for `SMEM` buffers.
    same_allocations = False
    if self.num_threads_per_block == 1:
      # When there is only one thread, we set `same_allocations` to `True`
      # regardless of whether `collective_axes` is set or not. Since the
      # allocation of barriers asserts on `same_allocations`, setting
      # `same_allocations = True` here ensures that barriers can be allocated
      # when only a single thread is present and `collective_axes` is empty.
      same_allocations = True
    elif collective_axes:
      if (
          self.mesh is None
          or len(collective_axes) != 1
          or collective_axes[0] != self.mesh.thread_name
      ):
        raise NotImplementedError(
            "When interpreting `run_scoped` in a GPU kernel, non-empty"
            " `collective_axes` is currently only supported when it contains a"
            " single axis that agrees with the thread axis (i.e. `thread_name`)"
            " of the mesh."
        )
      same_allocations = True

    # Allocate a buffer or barrier for each element of
    # `eqn.params['jaxpr'].invars`. It is assumed that each thread runs the same
    # sequence of `run_scoped`s.
    invars = eqn.params["jaxpr"].invars
    allocs = []
    for v in invars:
      allocs.append(_allocate_for_aval(v.aval, same_allocations))

    out = self.interpret(eqn.params["jaxpr"], *get_invals(), *allocs)

    for a, v in safe_zip(allocs, invars):
      _deallocate_for_aval(a, v.aval)

    return out

  def _interpret_cond_p(self, eqn, get_invals: Callable[[], Sequence[Any]]):
    assert eqn.primitive is lax.cond_p
    invals = get_invals()
    return lax.switch(
        invals[0],
        [
            functools.partial(self.interpret, branch_jaxpr.jaxpr)
            for branch_jaxpr in eqn.params["branches"]
        ],
        *invals[1:],
    )

  def _interpret_scan_p(self, eqn, get_invals: Callable[[], Sequence[Any]]):
    assert eqn.primitive is lax.scan_p
    consts, init_carry, xs = split_list(
        get_invals(),
        [eqn.params["num_consts"], eqn.params["num_carry"]],
    )

    def _scan_body(c, a):
      return split_list(
          self.interpret(eqn.params["jaxpr"].jaxpr, *consts, *c, *a),
          [eqn.params["num_carry"]],
      )

    carry, out = lax.scan(
        _scan_body, init_carry, xs=xs, length=eqn.params.get("length", None)
    )
    return carry + out

  def _interpret_barrier_primitive(
      self,
      eqn,
      get_invals: Callable[[], Sequence[Any]],
      barrier_callback: Callable[..., None],
  ):
    invals = get_invals()
    # `invals[0]` corresponds to the barrier this primitive operates on.
    allocation_key_as_array = _get_barrier_allocation_key_from_inval(
        invals[0], eqn.params["transforms_treedef"], invals[1:]
    )
    barrier_callback(
        device_id=self.device_info.device_id,
        thread_id=self.thread_id,
        allocation_key=allocation_key_as_array,
        source_info=eqn.source_info,
    )

    assert eqn.primitive.multiple_results
    return []

  def _interpret_barrier_arrive_p(
      self, eqn, get_invals: Callable[[], Sequence[Any]]
  ):
    assert eqn.primitive is gpu_primitives.barrier_arrive_p
    return self._interpret_barrier_primitive(
        eqn, get_invals, gpu_callbacks.call_barrier_arrive
    )

  def _interpret_barrier_wait_p(
      self, eqn, get_invals: Callable[[], Sequence[Any]]
  ):
    assert eqn.primitive is gpu_primitives.barrier_wait_p
    return self._interpret_barrier_primitive(
        eqn, get_invals, gpu_callbacks.call_barrier_wait
    )

  def _interpret_arithmetic_primitive(
      self, eqn, get_invals: Callable[[], Sequence[Any]]
  ):
    if self.interpret_params.skip_floating_point_ops and all(
        interpret_utils.is_float(ovar.aval.dtype) for ovar in eqn.outvars
    ):
      # Skip `eqn.primitive.bind` since `eqn.primitive` only produces
      # floating-point values. It is safe to populate `out` with avals
      # since mapping `env.write_many` over `out` (in `self.interpret`) below
      # only relies on the shape and dtype (for writing `Placeholder`s).
      out = [ovar.aval for ovar in eqn.outvars]
      if not eqn.primitive.multiple_results:
        out = out[0]
      return out
    else:
      bind_params = eqn.primitive.get_bind_params(eqn.params)
      return eqn.primitive.bind(*get_invals(), **bind_params)

  def _interpret_copy_gmem_to_smem_p(
      self, eqn, get_invals: Callable[[], Sequence[Any]]
  ):
    assert eqn.primitive is gpu_primitives.copy_gmem_to_smem_p
    invals = get_invals()

    # `invals[2]` corresponds to the barrier this primitive operates on.
    barrier_allocation_key_as_array = _get_barrier_allocation_key_from_inval(
        invals[2], eqn.params["barrier_transforms_treedef"], invals[3:]
    )

    gpu_callbacks.call_execute_device_local_memory_transfer(
        device_id=self.device_info.device_id,
        thread_id=self.thread_id,
        src_allocation_key=invals[0],
        src_transforms=(),
        dst_allocation_key=invals[1],
        dst_transforms=(),
        barrier_allocation_key_as_array=barrier_allocation_key_as_array,
        source_info=eqn.source_info,
    )
    return []

  def interpret(self, jaxpr, *args):
    sentinel_for_floating_point_values = (
        _SENTINEL if self.interpret_params.skip_floating_point_ops else None
    )
    env = interpret_utils.JaxprEnv(
        vars=jaxpr.constvars + jaxpr.invars,
        values=args,
        sentinel_for_floating_point_values=sentinel_for_floating_point_values,
    )

    for eqn in jaxpr.eqns:
      with source_info_util.user_context(
          eqn.source_info.traceback,
          name_stack=eqn.source_info.name_stack,
      ):
        # We defer reading the values for `eqn.invars` into each of the branches
        # of the match statement below. This is because the case for arithmetic
        # primitives may not need to do any reads
        # (if `self.interpret_params.skip_floating_point_ops` is True). If this
        # is the case, we want to avoid materializing the read array into the
        # jaxpr when this function is traced.
        deferred_invals = functools.partial(env.read_many, eqn.invars)
        match eqn.primitive:
          case lax.axis_index_p:
            out = self._interpret_axis_index_p(eqn)
          case primitives.program_id_p:
            # Currently we only support grids and clusters with a single device.
            # Hence, zero is the only valid program id.
            out = jnp.int32(0)
          case state_primitives.get_p:
            out = self._interpret_get_p(eqn, deferred_invals)
          case primitives.load_p:
            raise NotImplementedError("load_p is not supported on GPU yet")
          case state_primitives.swap_p:
            out = self._interpret_swap_p(eqn, deferred_invals)
          case primitives.swap_p:
            raise NotImplementedError("swap_p is not supported on GPU yet")
          case primitives.run_scoped_p:
            out = self._interpret_run_scoped_p(eqn, deferred_invals)
          case lax.cond_p:
            out = self._interpret_cond_p(eqn, deferred_invals)
          case lax.scan_p:
            out = self._interpret_scan_p(eqn, deferred_invals)
          case gpu_primitives.barrier_wait_p:
            out = self._interpret_barrier_wait_p(eqn, deferred_invals)
          case gpu_primitives.barrier_arrive_p:
            out = self._interpret_barrier_arrive_p(eqn, deferred_invals)
          case gpu_primitives.copy_gmem_to_smem_p:
            out = self._interpret_copy_gmem_to_smem_p(eqn, deferred_invals)
          case _:
            out = self._interpret_arithmetic_primitive(eqn, deferred_invals)

        out = out if eqn.primitive.multiple_results else [out]
        env.write_many(eqn.outvars, out)

    return env.read_many(jaxpr.outvars)



# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Custom fusible dtypes."""

import abc
import dataclasses
import functools
import itertools as it
from typing import Any, TypeVar
from collections.abc import Callable, Sequence

import jax
from jax._src import api_util
from jax._src import core
from jax._src import custom_derivatives
from jax._src import dtypes
from jax._src import linear_util as lu
from jax._src import source_info_util
from jax._src import state
from jax._src import tree_util
from jax._src import util
from jax._src.interpreters import partial_eval as pe
from jax._src.lax.control_flow import conditionals
from jax._src.pallas import core as pallas_core
from jax._src.pallas import pallas_call
from jax._src.pallas import primitives as pallas_primitives
from jax._src.pallas.fuser import block_spec
from jax._src.pallas.fuser.fusible import fusible_p
from jax._src.state import discharge as state_discharge
from jax._src.state import primitives as state_primitives
from jax._src.util import foreach

map, unsafe_map = util.safe_map, map
zip, unsafe_zip = util.safe_zip, zip

T = TypeVar("T")

_physicalize_rules: dict[core.Primitive, Callable[..., Any]] = {}

pack_dtype_p = core.Primitive("pack_dtype")


@pack_dtype_p.def_abstract_eval
def pack_dtype_abstract_eval(*xs, dtype):
  if dtypes.issubdtype(dtype, FusibleElementDType):
    return dtype.abstract_pack(*xs)
  raise ValueError("Attempted to pack non-fusion dtype: {dtype}")


def pack(*xs, dtype):
  return pack_dtype_p.bind(*xs, dtype=dtype)


unpack_dtype_p = core.Primitive("unpack_dtype")
unpack_dtype_p.multiple_results = True


@unpack_dtype_p.def_abstract_eval
def unpack_dtype_abstract_eval(x):
  if dtypes.issubdtype(x.dtype, FusibleElementDType):
    return x.dtype.abstract_unpack(x)
  elif isinstance(x.dtype, state.AbstractRef):
    raise NotImplementedError()
  raise ValueError("Attempted to unpack non-fusion dtype: {dtype}")


def unpack(x):
  return tuple(unpack_dtype_p.bind(x))


class FusibleElementDType(dtypes.extended):
  """Scalar dtype for fusible dtypes."""


class FusibleTyRules:
  allow_conversion: bool = False


class FusionDType(dtypes.ExtendedDType, util.StrictABC):
  """Base class for fusible extended dtypes."""

  _op_registry = {}
  _rules = FusibleTyRules
  type = FusibleElementDType

  @abc.abstractmethod
  def abstract_unpack(self, x) -> Sequence[Any]:
    raise NotImplementedError()

  @abc.abstractmethod
  def abstract_pack(self, *xs):
    raise NotImplementedError()

  @classmethod
  def register_op(cls, primitive):
    def _register_fn(fn):
      cls._op_registry[primitive] = fn

    return _register_fn

  @classmethod
  def get_op_rule(cls, primitive):
    return cls._op_registry.get(primitive)

  @property
  def name(self):
    return str(self)

  @abc.abstractmethod
  def pull_block_spec_one_step(self, aval_out, *args, **kwargs):
    raise NotImplementedError()

  @abc.abstractmethod
  def unpack_push_block_spec(self, aval_in, *args, **kwargs):
    raise NotImplementedError()

  @abc.abstractmethod
  def unpack_pull_block_spec(self, aval_in, *args, **kwargs):
    raise NotImplementedError()


def physicalize(f):
  """Runs a function that contains fusible extended dtypes."""

  def wrapper(*args, **kwargs):
    if kwargs:
      raise NotImplementedError()
    flattened_args, treedef = jax.tree.flatten(args)
    debug_info = api_util.debug_info("physicalize", f, args, kwargs)
    wrapped_fun, out_tree_thunk = api_util.flatten_fun_nokwargs(
        lu.wrap_init(f, debug_info=debug_info), treedef
    )
    avals = [core.typeof(a) for a in flattened_args]
    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, avals)
    new_jaxpr = physicalize_closed_jaxpr(core.ClosedJaxpr(jaxpr, consts))
    out_flat = core.eval_jaxpr(
        new_jaxpr.jaxpr, new_jaxpr.consts, *flattened_args
    )
    return tree_util.tree_unflatten(out_tree_thunk(), out_flat)

  return wrapper


@util.weakref_lru_cache
def physicalize_closed_jaxpr(jaxpr: core.ClosedJaxpr) -> core.ClosedJaxpr:
  """Replaces all extended dtypes with physical types in a jaxpr."""
  fun = functools.partial(physicalize_interp, jaxpr.jaxpr, jaxpr.consts)
  in_avals = [_physical_aval(aval) for aval in jaxpr.in_avals]
  flat_avals, treedef = tree_util.tree_flatten(in_avals)
  debug_info = api_util.debug_info("physicalize_closed_jaxpr", fun, (), {})
  wrapped_fun, _ = api_util.flatten_fun_nokwargs(
      lu.wrap_init(fun, debug_info=debug_info.with_unknown_names()), treedef
  )
  new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, flat_avals)
  assert len(new_jaxpr.constvars) == len(consts), "Mismatched consts"
  return core.ClosedJaxpr(new_jaxpr, consts)


def _physical_aval(aval):
  if isinstance(aval, core.ShapedArray):
    if isinstance(aval.dtype, FusionDType):
      return aval.dtype.abstract_unpack(aval)
    return core.ShapedArray(aval.shape, aval.dtype)
  if isinstance(aval, state.AbstractRef):
    if _is_fusion_type(aval):
      unpacked = aval.dtype.abstract_unpack(aval.inner_aval)
      return tuple(aval.update(inner_aval=u) for u in unpacked)
    return aval
  return aval


def physicalize_jaxpr(jaxpr: core.Jaxpr) -> core.Jaxpr:
  """Replaces all extended dtypes with physical types in a jaxpr."""

  def _flat_jaxpr_eval(consts, args):
    return physicalize_interp(jaxpr, consts, *args)

  in_avals = [_physical_aval(v.aval) for v in jaxpr.invars]
  const_avals = [_physical_aval(v.aval) for v in jaxpr.constvars]
  flat_avals, treedef = jax.tree.flatten((const_avals, in_avals))
  debug_info = api_util.debug_info(
      "physicalize_jaxpr", _flat_jaxpr_eval, (const_avals, in_avals), {})
  wrapped_fun, _ = api_util.flatten_fun_nokwargs(
      lu.wrap_init(_flat_jaxpr_eval, debug_info=debug_info), treedef
  )
  new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, flat_avals)
  assert not consts
  new_jaxpr = pe.convert_invars_to_constvars(
      new_jaxpr, len(tree_util.tree_leaves(const_avals))
  )
  return new_jaxpr


@dataclasses.dataclass
class Context:
  avals_in: Sequence[Any]
  avals_out: Sequence[Any]


def physicalize_interp(
    jaxpr: core.Jaxpr, consts: Sequence[core.Value], *args: core.Value
):
  """Physicalizes a jaxpr by replacing fusible dtypes with physical types."""
  # TODO: Merge into JAX core.
  env: dict[core.Var, Any] = {}

  def read_env(var: core.Atom):
    if isinstance(var, core.Literal):
      return var.val
    return env[var]

  def write_env(var: core.Var, val: Any):
    env[var] = val

  foreach(write_env, jaxpr.constvars, consts)
  assert len(jaxpr.invars) == len(
      args
  ), f"Length mismatch: {jaxpr.invars} != {args}"
  foreach(write_env, jaxpr.invars, args)

  for eqn in jaxpr.eqns:
    invals = list(map(read_env, eqn.invars))
    avals_in = tuple(x.aval for x in eqn.invars)
    name_stack = (
        source_info_util.current_name_stack() + eqn.source_info.name_stack
    )
    with (
        source_info_util.user_context(
            eqn.source_info.traceback, name_stack=name_stack
        ),
        eqn.ctx.manager,
    ):
      # need to check types and then invoke the correct rule.
      ctx = Context(
          avals_in=avals_in, avals_out=[var.aval for var in eqn.outvars]
      )
      custom_rule = _phys_find_rule(eqn.primitive, avals_in)
      if custom_rule:
        outvals = custom_rule(ctx, *invals, **eqn.params)
      else:
        bind_params = eqn.primitive.get_bind_params(eqn.params)
        outvals = eqn.primitive.bind(*invals, **bind_params)

    if eqn.primitive.multiple_results:
      assert len(outvals) == len(eqn.outvars), eqn
      foreach(write_env, eqn.outvars, outvals)
    else:
      write_env(eqn.outvars[0], outvals)

  return map(read_env, jaxpr.outvars)


def _is_fusion_type(aval: core.AbstractValue):
  """Returns whether an aval is an array containing fusion types."""
  return (
      isinstance(aval, (core.ShapedArray, state.AbstractRef))
      and hasattr(aval, 'dtype')
      and isinstance(aval.dtype, FusionDType)
  )


def _phys_find_rule(primitive, avals: Sequence[core.AbstractValue]):
  """Finds the physicalization rule for a primitive."""
  if primitive in _physicalize_rules:
    return _physicalize_rules[primitive]

  # pyrefly: ignore[missing-attribute]
  fusion_types = {aval.dtype for aval in avals if _is_fusion_type(aval)}
  if len(fusion_types) == 0:
    return None
  elif len(fusion_types) > 1:
    raise ValueError(f"Multiple fusion types for primitive: {fusion_types}")
  fusion_type = fusion_types.pop()
  if primitive not in fusion_type._op_registry:
    raise ValueError(
        f"No implementation found for primitive {primitive} "
        f"for custom type {fusion_type}"
    )
  return fusion_type.get_op_rule(primitive)


def _assert_no_fusion_types(avals: Sequence[core.AbstractValue]):
  if any(_is_fusion_type(aval) for aval in avals):
    raise NotImplementedError(f"Fusion type found in avals: {avals}")


def _pallas_call_physicalize_rule(
    ctx: Context, *args, jaxpr, grid_mapping: pallas_core.GridMapping, **kwargs
):
  _assert_no_fusion_types(ctx.avals_in)
  _assert_no_fusion_types(ctx.avals_out)
  with grid_mapping.trace_env():
    new_jaxpr = physicalize_closed_jaxpr(core.ClosedJaxpr(jaxpr, ()))
  if diff := len(new_jaxpr.jaxpr.invars) - len(jaxpr.invars):
    num_scratch_avals = len(grid_mapping.scratch_avals) + diff
    new_scratch_avals = tuple(v.aval for v in
                              new_jaxpr.jaxpr.invars[-num_scratch_avals:])
    grid_mapping = grid_mapping.replace(
        scratch_avals=new_scratch_avals
    )
  return pallas_call.pallas_call_p.bind(
      *args, jaxpr=new_jaxpr.jaxpr, grid_mapping=grid_mapping, **kwargs
  )


_physicalize_rules[pallas_call.pallas_call_p] = _pallas_call_physicalize_rule


def _cond_physicalize_rule(ctx: Context, *args, branches, **kwargs):
  _assert_no_fusion_types(ctx.avals_out)
  physicalized_branches = tuple(
      physicalize_closed_jaxpr(branch) for branch in branches
  )
  flat_args = jax.tree.leaves(args)
  return conditionals.cond_p.bind(
      *flat_args, branches=physicalized_branches, **kwargs
  )


_physicalize_rules[conditionals.cond_p] = _cond_physicalize_rule


@lu.transformation2
def _physicalize_transform(f, *args):
  vals, zeros = args[::2], args[1::2]
  assert len(vals) == len(zeros)
  wrapper = lambda *inner_vals: f(
      *it.chain.from_iterable(zip(inner_vals, zeros))
  )
  return physicalize(wrapper)(*vals)


@lu.transformation2
def _physicalize_transform_bwd(f, const_avals, *args):
  return [custom_derivatives.Zero(a) for a in const_avals] + list(
      physicalize(f)(*args)
  )


def _custom_vjp_call_physicalize_rule(
    ctx: Context, *args, call_jaxpr, num_consts, fwd_jaxpr_thunk, bwd, **kwargs
):
  _assert_no_fusion_types(ctx.avals_out)
  new_jaxpr = physicalize_closed_jaxpr(call_jaxpr)
  fun = lu.wrap_init(core.jaxpr_as_fun(new_jaxpr),
                     debug_info=call_jaxpr.jaxpr.debug_info)
  fwd = custom_derivatives.lift_fwd(num_consts, fwd_jaxpr_thunk)
  fwd_physicalized = _physicalize_transform(fwd)
  const_avals, _ = util.split_list(new_jaxpr.in_avals, [num_consts])
  bwd_physicalized = _physicalize_transform_bwd(bwd, const_avals)
  kwargs['subfuns'] = (fun, fwd_physicalized, bwd_physicalized)
  return custom_derivatives.custom_vjp_call_p.bind(*args, **kwargs)

_physicalize_rules[custom_derivatives.custom_vjp_call_p] = _custom_vjp_call_physicalize_rule


def _run_state_rule(ctx: Context, *args, jaxpr, which_linear, is_initialized):
  _assert_no_fusion_types(ctx.avals_in)
  _assert_no_fusion_types(ctx.avals_out)
  jaxpr = physicalize_jaxpr(jaxpr)
  return state_discharge.run_state_p.bind(
      *args,
      jaxpr=jaxpr,
      which_linear=which_linear,
      is_initialized=is_initialized,
  )


_physicalize_rules[state_discharge.run_state_p] = _run_state_rule


def _core_map_rule(ctx: Context, *args, jaxpr, **params):
  _assert_no_fusion_types(ctx.avals_in)
  _assert_no_fusion_types(ctx.avals_out)
  assert not jaxpr.invars
  with core.extend_axis_env_nd(params["mesh"].shape.items()):
    jaxpr = physicalize_jaxpr(jaxpr)
  return pallas_core.core_map_p.bind(*args, jaxpr=jaxpr, **params)


_physicalize_rules[pallas_core.core_map_p] = _core_map_rule


def _run_scoped_rule(ctx: Context, *args, jaxpr, **params):
  _assert_no_fusion_types(ctx.avals_out)
  jaxpr = physicalize_jaxpr(jaxpr)
  flat_args = tree_util.tree_leaves(args)
  assert len(flat_args) == len(
      jaxpr.constvars
  ), f"Length mismatch: {len(flat_args)=} != {len(jaxpr.constvars)=}"
  return pallas_primitives.run_scoped_p.bind(*flat_args, jaxpr=jaxpr, **params)


_physicalize_rules[pallas_primitives.run_scoped_p] = _run_scoped_rule


def _scan_rule(ctx: Context, *args, jaxpr, **params):
  _assert_no_fusion_types(ctx.avals_in)
  _assert_no_fusion_types(ctx.avals_out)
  jaxpr = physicalize_closed_jaxpr(jaxpr)
  return jax.lax.scan_p.bind(*args, jaxpr=jaxpr, **params)


_physicalize_rules[jax.lax.scan_p] = _scan_rule


def _while_rule(
    ctx: Context, *args, body_jaxpr, cond_jaxpr, body_nconsts,
    cond_nconsts, **params
):
  _assert_no_fusion_types(ctx.avals_out)
  cond_avals = [v.aval for v in cond_jaxpr.jaxpr.invars]
  _, cond_in_avals = util.split_list(cond_avals, [cond_nconsts])
  _assert_no_fusion_types(cond_in_avals)
  new_cond_jaxpr = physicalize_closed_jaxpr(cond_jaxpr)
  new_num_cond_consts = (
      cond_nconsts
      + len(new_cond_jaxpr.jaxpr.invars)
      - len(cond_jaxpr.jaxpr.invars)
  )

  body_avals = [v.aval for v in body_jaxpr.jaxpr.invars]
  _, body_in_avals = util.split_list(body_avals, [body_nconsts])
  _assert_no_fusion_types(body_in_avals)
  new_body_jaxpr = physicalize_closed_jaxpr(body_jaxpr)
  new_num_body_consts = (
      body_nconsts
      + len(new_body_jaxpr.jaxpr.invars)
      - len(body_jaxpr.jaxpr.invars)
  )
  flat_args = tree_util.tree_leaves(args)
  cond_consts, body_consts, flat_args = util.split_list(
      flat_args, [new_num_cond_consts, new_num_body_consts]
  )
  assert len(flat_args) + len(body_consts) == len(
      new_body_jaxpr.jaxpr.invars), (
      f"Length mismatch: {len(flat_args) + len(body_consts)} !="
      f" {len(new_body_jaxpr.jaxpr.invars)=}"
  )
  assert len(flat_args) + len(cond_consts) == len(
      new_cond_jaxpr.jaxpr.invars), (
      f"Length mismatch: {len(flat_args) + len(cond_consts)} !="
      f" {len(new_cond_jaxpr.jaxpr.invars)=}"
  )
  return jax.lax.while_p.bind(
      *(cond_consts + body_consts + flat_args),
      body_jaxpr=new_body_jaxpr,
      cond_jaxpr=new_cond_jaxpr,
      body_nconsts=new_num_body_consts,
      cond_nconsts=new_num_cond_consts,
      **params,
  )


_physicalize_rules[jax.lax.while_p] = _while_rule


def _pack_rule(_, *args, dtype):
  del dtype
  return args


_physicalize_rules[pack_dtype_p] = _pack_rule


def _unpack_rule(_, arg):
  return arg


_physicalize_rules[unpack_dtype_p] = _unpack_rule


def _swap_rule(ctx: Context, ref, val, *args, tree):
  ref_aval, *_ = ctx.avals_in
  if not _is_fusion_type(ref_aval):
    return state_primitives.swap_p.bind(ref, val, *args, tree=tree)
  return ref_aval.dtype.swap(ref, val, *args, tree=tree)


_physicalize_rules[state_primitives.swap_p] = _swap_rule


def _get_rule(ctx: Context, ref, *args, tree):
  ref_aval, *_ = ctx.avals_in
  if not _is_fusion_type(ref_aval):
    return state_primitives.get_p.bind(ref, *args, tree=tree)
  return ref_aval.dtype.get(ref, *args, tree=tree)


_physicalize_rules[state_primitives.get_p] = _get_rule


@block_spec.register_eval_rule(pack_dtype_p)
def _pack_dtype_eval_rule(ctx: block_spec.KernelEvalContext, *args, dtype):
  return dtype.pack_eval_rule(ctx, *args)


@block_spec.register_pull_block_spec_rule(pack_dtype_p)
def _pack_dtype_pull_rule(
    ctx: block_spec.PullRuleContext,
    block_spec: pallas_core.BlockSpec,
    *,
    dtype: FusionDType,
):
  aval_out = ctx.avals_out[0]
  return dtype.pull_block_spec_one_step(aval_out, block_spec)


@block_spec.register_push_block_spec_rule(unpack_dtype_p)
def _unpack_dtype_push_rule(
    ctx: block_spec.PushRuleContext,
    block_spec: pallas_core.BlockSpec,
):
  aval_in = ctx.avals_in[0]
  assert isinstance(aval_in, core.ShapedArray)
  assert isinstance(aval_in.dtype, FusionDType), aval_in.dtype
  return aval_in.dtype.unpack_push_block_spec(aval_in, block_spec)


@block_spec.register_pull_block_spec_rule(unpack_dtype_p)
def _unpack_dtype_pull_rule(
    ctx: block_spec.PushRuleContext,
    block_specs: pallas_core.BlockSpec,
):
  aval_in = ctx.avals_in[0]
  assert isinstance(aval_in, core.ShapedArray)
  assert isinstance(aval_in.dtype, FusionDType), aval_in.dtype
  return aval_in.dtype.unpack_pull_block_spec(aval_in, *block_specs)  # pyrefly: ignore[not-iterable]


@block_spec.register_eval_rule(unpack_dtype_p)
def _unpack_dtype_eval_rule(ctx: block_spec.KernelEvalContext, *args):
  assert ctx.avals_in is not None
  aval_in = ctx.avals_in[0]
  assert isinstance(aval_in, core.ShapedArray)
  assert isinstance(aval_in.dtype, FusionDType), aval_in.dtype
  return aval_in.dtype.unpack_eval_rule(ctx, *args)  # pyrefly: ignore[missing-attribute]


def _fusible_physicalize_rule(
    _, *consts_and_args, jaxpr, num_consts, in_tree, out_tree, func
):
  consts, _ = util.split_list(consts_and_args, [num_consts])
  new_jaxpr = physicalize_closed_jaxpr(core.ClosedJaxpr(jaxpr, consts))
  return fusible_p.bind(
      *consts_and_args,
      jaxpr=new_jaxpr.jaxpr,
      num_consts=num_consts,
      in_tree=in_tree,
      out_tree=out_tree,
      func=func,
  )


_physicalize_rules[fusible_p] = _fusible_physicalize_rule



# Copyright 2022 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/jax-ml/jax/issues/7570

import jax._src.core as _src_core

from jax._src.core import (
  AbstractValue as AbstractValue,
  Atom as Atom,
  ParamDict as ParamDict,
  ShapedArray as ShapedArray,
  Trace as Trace,
  Tracer as Tracer,
  Value as Value,
  ensure_compile_time_eval as ensure_compile_time_eval,
  eval_context as eval_context,
  eval_jaxpr as eval_jaxpr,
  max_dim as max_dim,
  min_dim as min_dim,
)

_deprecations = {
  # Deprecated in v0.8.2; finalized in v0.10.0.
  # TODO(jakevdp) remove entries in v0.11.0.
  "get_aval": (
    "jax.core.get_aval was deprecated in JAX v0.8.2 and removed in JAX v0.10.0;"
    " use jax.typeof instead.",
    None,
  ),
  "mapped_aval": (
    "jax.core.mapped_aval was deprecated in JAX v0.8.2 and removed in JAX"
    " v0.10.0. Use jax.extend.core.mapped_aval.",
    None,
  ),
  "unmapped_aval": (
    "jax.core.unmapped_aval was deprecated in JAX v0.8.2 and removed in JAX"
    " v0.10.0. Use jax.extend.core.unmapped_aval.",
    None,
  ),
  "set_current_trace": (
    "jax.core.set_current_trace was deprecated in JAX v0.8.2 and removed in"
    " JAX v0.10.0. Use jax.extend.core.set_current_trace.",
    None,
  ),
  "take_current_trace": (
    "jax.core.take_current_trace was deprecated in JAX v0.8.2 and removed in"
    " JAX v0.10.0. Use jax.extend.core.take_current_trace.",
    None,
  ),
  "traverse_jaxpr_params": (
    "jax.core.traverse_jaxpr_params was deprecated in JAX v0.8.2 and removed in"
    " JAX v0.10.0.",
    None,
  ),
  "TraceTag": (
    "jax.core.TraceTag was deprecated in JAX v0.8.2 and removed in JAX v0.10.0."
    " Use jax.extend.core.TraceTag.",
    None,
  ),
  "call_impl": (
    "jax.core.call_impl was deprecated in JAX v0.8.2 and removed in JAX"
    " v0.10.0. Use jax.extend.core.call_impl.",
    None,
  ),
  "subjaxprs": (
    "jax.core.subjaxprs was deprecated in JAX v0.8.2 and removed in JAX"
    " v0.10.0. Use jax.extend.core.subjaxprs.",
    None,
  ),
  "AbstractToken": (
    "jax.core.AbstractToken was deprecated in JAX v0.8.2 and removed in JAX"
    " v0.10.0. Use jax.extend.core.AbstractToken.",
    None,
  ),
  # Deprecated in JAX v0.10.0; TODO(jakevdp) finalize in v0.11.0
  "CallPrimitive": (
    "jax.core.CallPrimitive is deprecated. Use jax.extend.core.CallPrimitive.",
    _src_core.CallPrimitive,
  ),
  "DebugInfo": (
    "jax.core.DebugInfo is deprecated. Use jax.extend.core.DebugInfo.",
    _src_core.DebugInfo,
  ),
  "DropVar": (
    "jax.core.DropVar is deprecated. Use jax.extend.core.DropVar.",
    _src_core.DropVar,
  ),
  "Effect": (
    "jax.core.Effect is deprecated. Use jax.extend.core.Effect.",
    _src_core.Effect,
  ),
  "Effects": (
    "jax.core.Effects is deprecated. Use jax.extend.core.Effects.",
    _src_core.Effects,
  ),
  "InconclusiveDimensionOperation": (
    "jax.core.InconclusiveDimensionOperation is deprecated. Use jax.extend.core.InconclusiveDimensionOperation.",
    _src_core.InconclusiveDimensionOperation,
  ),
  "JaxprTypeError": (
    "jax.core.JaxprTypeError is deprecated. Use jax.extend.core.JaxprTypeError.",
    _src_core.JaxprTypeError,
  ),
  "check_jaxpr": (
    "jax.core.check_jaxpr is deprecated. Use jax.extend.core.check_jaxpr.",
    _src_core.check_jaxpr,
  ),
  "concrete_or_error": (
    "jax.core.concrete_or_error is deprecated. Use jax.extend.core.concrete_or_error.",
    _src_core.concrete_or_error,
  ),
  "find_top_trace": (
    "jax.core.find_top_trace is deprecated. Use jax.extend.core.find_top_trace.",
    _src_core.find_top_trace,
  ),
  "gensym": (
    "jax.core.gensym is deprecated. Use jax.extend.core.gensym.",
    _src_core.gensym,
  ),
  "get_opaque_trace_state": (
    "jax.core.get_opaque_trace_state is deprecated. Use jax.extend.core.get_opaque_trace_state.",
    _src_core.get_opaque_trace_state,
  ),
  "jaxprs_in_params": (
    "jax.core.jaxprs_in_params is deprecated. Use jax.extend.core.jaxprs_in_params.",
    _src_core.jaxprs_in_params,
  ),
  "new_jaxpr_eqn": (
    "jax.core.new_jaxpr_eqn is deprecated. Use jax.extend.core.new_jaxpr_eqn.",
    _src_core.new_jaxpr_eqn,
  ),
  "no_effects": (
    "jax.core.no_effects is deprecated. Use jax.extend.core.no_effects.",
    _src_core.no_effects,
  ),
  "nonempty_axis_env_DO_NOT_USE": (
    "jax.core.nonempty_axis_env_DO_NOT_USE is deprecated.",
    _src_core.nonempty_axis_env,
  ),
  "primal_dtype_to_tangent_dtype": (
    "jax.core.primal_dtype_to_tangent_dtype is deprecated. Use jax.extend.core.primal_dtype_to_tangent_dtype.",
    _src_core.primal_dtype_to_tangent_dtype,
  ),
  "unsafe_am_i_under_a_jit_DO_NOT_USE": (
    "jax.core.unsafe_am_i_under_a_jit_DO_NOT_USE is deprecated.",
    _src_core.unsafe_am_i_under_a_jit,
  ),
  "unsafe_am_i_under_a_vmap_DO_NOT_USE": (
    "jax.core.unsafe_am_i_under_a_vmap_DO_NOT_USE is deprecated.",
    _src_core.unsafe_am_i_under_a_vmap,
  ),
  "unsafe_get_axis_names_DO_NOT_USE": (
    "jax.core.unsafe_get_axis_names_DO_NOT_USE is deprecated.",
    _src_core.unsafe_get_axis_names,
  ),
  "valid_jaxtype": (
    "jax.core.valid_jaxtype is deprecated. Use jax.extend.core.valid_jaxtype.",
    _src_core.valid_jaxtype,
  ),
  "JaxprPpContext": (
    "jax.core.JaxprPpContext is deprecated.",
    _src_core.JaxprPpContext,
  ),
  "JaxprPpSettings": (
    "jax.core.JaxprPpSettings is deprecated.",
    _src_core.JaxprPpSettings,
  ),
  "OutputType": (
    "jax.core.OutputType is deprecated.",
    _src_core.OutputType,
  ),
  "abstract_token": (
    "jax.core.abstract_token is deprecated.",
    _src_core.abstract_token,
  ),
  "aval_mapping_handlers": (
    "jax.core.aval_mapping_handlers is deprecated.",
    _src_core.aval_mapping_handlers,
  ),
  "call": (
    "jax.core.call is deprecated.",
    _src_core.call,
  ),
  "concretization_function_error": (
    "jax.core.concretization_function_error is deprecated.",
    _src_core.concretization_function_error,
  ),
  "custom_typechecks": (
    "jax.core.custom_typechecks is deprecated.",
    _src_core.custom_typechecks,
  ),
  "is_concrete": (
    "jax.core.is_concrete is deprecated.",
    _src_core.is_concrete,
  ),
  "is_constant_dim": (
    "jax.core.is_constant_dim is deprecated.",
    _src_core.is_constant_dim,
  ),
  "is_constant_shape": (
    "jax.core.is_constant_shape is deprecated.",
    _src_core.is_constant_shape,
  ),
  "literalable_types": (
    "jax.core.literalable_types is deprecated.",
    _src_core.literalable_types,
  ),
  "no_axis_name": (
    "jax.core.no_axis_name is deprecated.",
    _src_core.no_axis_name,
  ),
  "pytype_aval_mappings": (
    "jax.core.pytype_aval_mappings is deprecated.",
    _src_core.pytype_aval_mappings,
  ),
  "trace_ctx": (
    "jax.core.trace_ctx is deprecated.",
    _src_core.trace_ctx,
  ),
}

import typing as _typing
if _typing.TYPE_CHECKING:
  CallPrimitive = _src_core.CallPrimitive
  DebugInfo = _src_core.DebugInfo
  DropVar = _src_core.DropVar
  Effect = _src_core.Effect
  Effects = _src_core.Effects
  InconclusiveDimensionOperation = _src_core.InconclusiveDimensionOperation
  JaxprPpContext = _src_core.JaxprPpContext
  JaxprPpSettings = _src_core.JaxprPpSettings
  JaxprTypeError = _src_core.JaxprTypeError
  OutputType = _src_core.OutputType
  abstract_token = _src_core.abstract_token
  aval_mapping_handlers = _src_core.aval_mapping_handlers
  call = _src_core.call
  check_jaxpr = _src_core.check_jaxpr
  concrete_or_error = _src_core.concrete_or_error
  concretization_function_error = _src_core.concretization_function_error
  custom_typechecks = _src_core.custom_typechecks
  find_top_trace = _src_core.find_top_trace
  gensym = _src_core.gensym
  get_opaque_trace_state = _src_core.get_opaque_trace_state
  is_concrete = _src_core.is_concrete
  is_constant_dim = _src_core.is_constant_dim
  is_constant_shape = _src_core.is_constant_shape
  jaxprs_in_params = _src_core.jaxprs_in_params
  literalable_types = _src_core.literalable_types
  new_jaxpr_eqn = _src_core.new_jaxpr_eqn
  no_axis_name = _src_core.no_axis_name
  no_effects = _src_core.no_effects
  nonempty_axis_env_DO_NOT_USE = _src_core.nonempty_axis_env
  primal_dtype_to_tangent_dtype = _src_core.primal_dtype_to_tangent_dtype
  pytype_aval_mappings = _src_core.pytype_aval_mappings
  trace_ctx = _src_core.trace_ctx
  unsafe_am_i_under_a_jit_DO_NOT_USE = _src_core.unsafe_am_i_under_a_jit
  unsafe_am_i_under_a_vmap_DO_NOT_USE = _src_core.unsafe_am_i_under_a_vmap
  unsafe_get_axis_names_DO_NOT_USE = _src_core.unsafe_get_axis_names
  valid_jaxtype = _src_core.valid_jaxtype
else:
  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
  __getattr__ = _deprecation_getattr(__name__, _deprecations)
  del _deprecation_getattr
del _typing
del _src_core
